"""abstract base for python-based study executors"""

import abc
import pathlib
import re
import sys
import typing

from cumulus_library import base_utils, study_manifest


class BaseTableBuilder(abc.ABC):
    """Generic base class for python table builders.

    To use a table builder, extend this class exactly once in a new module.
    See ./studies/vocab or ./studies/core for example usage.
    """

    display_text = "Building custom tables..."

    def __init__(self):
        self.queries = []

    @abc.abstractmethod
    def prepare_queries(
        self,
        config: base_utils.StudyConfig,
        manifest: study_manifest.StudyManifest,
        *args,
        **kwargs,
    ):
        """Main entrypoint for python table builders.

        When completed, prepare_queries should populate self.queries with sql
        statements to execute. This array will the be read by execute_queries.

        :param config: A study configuration object
        """
        raise NotImplementedError  # pragma: no cover

    @typing.final
    def execute_queries(
        self,
        config: base_utils.StudyConfig,
        manifest: study_manifest.StudyManifest,
        *args,
        **kwargs,
    ):
        """Executes queries set up by a prepare_queries call

        :param config: A study configuration object
        """
        self.prepare_queries(*args, config=config, manifest=manifest, **kwargs)
        cursor = config.db.cursor()
        if config.drop_table:
            table_names = []
            for query in self.queries:
                # Get the first non-whitespace word after create table
                table_name = re.search('(?i)(?<=create table )(([a-zA-Z0-9_".-]+))', query)

                if table_name:
                    if table_name[0] == "IF":
                        # Edge case - if we're doing an empty conditional CTAS creation,
                        # we need to run a slightly different regex
                        table_name = re.search('(?i)(?<=not exists )(([a-zA-Z0-9_".-]+))', query)

                    table_name = table_name[0]
                    table_names.append(table_name)
            for table_name in table_names:
                cursor.execute(f"DROP TABLE IF EXISTS {table_name}")
        with base_utils.get_progress_bar(disable=config.verbose) as progress:
            task = progress.add_task(
                self.display_text,
                total=len(self.queries),
                visible=not config.verbose,
            )
            for query in self.queries:
                try:
                    query = base_utils.update_query_if_schema_specified(query, manifest)
                    with base_utils.query_console_output(config.verbose, query, progress, task):
                        cursor.execute(query)
                except Exception as e:  # pylint: disable=broad-exception-caught
                    sys.exit(
                        "An error occured executing this query:\n----\n" f"{query}\n----\n" f"{e}"
                    )

        self.post_execution(config, *args, **kwargs)

    def post_execution(
        self,
        config: base_utils.StudyConfig,
        *args,
        **kwargs,
    ):
        """Hook for any additional actions to run after execute_queries"""
        pass

    def comment_queries(self, doc_str=None):
        """Convenience method for annotating outputs of template generators to disk"""
        commented_queries = ["-- noqa: disable=all"]
        if doc_str:
            commented_queries.append(doc_str)
            commented_queries.append(
                "\n-- ###########################################################\n"
            )
        for query in self.queries:
            commented_queries.append(query)
            commented_queries.append(
                "\n-- ###########################################################\n"
            )
        commented_queries.pop()
        self.queries = commented_queries

    def write_queries(self, path: pathlib.Path | None = None):
        """writes all queries constructed by prepare_queries to disk"""
        if path is None:
            path = pathlib.Path.cwd() / "output.sql"  # pragma: no cover

        path.parents[0].mkdir(parents=True, exist_ok=True)
        with open(path, "w", encoding="utf-8") as file:
            for query in self.queries:
                file.write(query)
                file.write("\n")



"""Module for generating encounter codeableConcept table"""

import cumulus_library
from cumulus_library import base_utils
from cumulus_library.studies.discovery import code_definitions
from cumulus_library.studies.discovery.discovery_templates import discovery_templates
from cumulus_library.template_sql import sql_utils


class CodeDetectionBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Selecting unique code systems..."

    def _check_coding_against_db(self, code_source, database):
        """selects the appropriate DB query to run"""

        return sql_utils.is_field_populated(
            database=database,
            source_table=code_source["table_name"],
            hierarchy=code_source["column_hierarchy"],
            expected=code_source.get("expected"),
        )

    def _check_codes_in_fields(self, code_sources: list[dict], database) -> dict:
        """checks if Coding/CodeableConcept fields are present and populated"""

        with base_utils.get_progress_bar() as progress:
            task = progress.add_task(
                "Discovering available coding systems...",
                total=len(code_sources),
            )
            for code_source in code_sources:
                code_source["has_data"] = self._check_coding_against_db(code_source, database)
                progress.advance(task)
        return code_sources

    def prepare_queries(
        self,
        *args,
        config: cumulus_library.StudyConfig,
        **kwargs,
    ):
        """Constructs queries related to condition codeableConcept

        :param config: A study config object
        """

        code_sources = []
        required_keys = {"table_name", "column_hierarchy"}
        for code_definition in code_definitions.code_list:
            if not required_keys.issubset(code_definition):
                raise KeyError(
                    "Expected table_name and column_hierarchy keys in " f"{code_definition!s}"
                )
            code_source = {
                "has_data": False,
            }
            for key in code_definition.keys():
                code_source[key] = code_definition[key]
            code_sources.append(code_source)
        code_sources = self._check_codes_in_fields(code_sources, config.db)
        query = discovery_templates.get_system_pairs("discovery__code_sources", code_sources)
        self.queries.append(query)



import dataclasses


@dataclasses.dataclass(kw_only=True)
class ValuesetConfig:
    """Provides expected values for creating valuesets"""

    rules_file: str = None
    keyword_file: str = None
    table_prefix: str = None
    umls_stewards: dict[str, str] = None
    vsac_stewards: dict[str, str] = None



"""Package metadata"""

from cumulus_library.base_utils import StudyConfig
from cumulus_library.builders.base_table_builder import BaseTableBuilder
from cumulus_library.builders.counts import CountsBuilder
from cumulus_library.study_manifest import StudyManifest

__all__ = ["BaseTableBuilder", "CountsBuilder", "StudyConfig", "StudyManifest"]
__version__ = "4.1.2"



from dataclasses import dataclass
from enum import Enum
from pathlib import Path

from cumulus_library.errors import CountsBuilderError
from cumulus_library.template_sql import base_templates


class CountableFhirResource(Enum):
    """Contains FHIR types for which we have count table generation support.

    This is primarily used as a way to decide if a secondary join is needed to get the
    appropriate ID field to count. Patient (and a value of None) skip this check,
    while the other resources will add an additional source to get the proper countable
    ids.
    """

    ALLERGYINTOLERANCE = "allergyintolerance"
    CONDITION = "condition"
    DIAGNOSTICREPORT = "diagnosticreport"
    DOCUMENTREFERENCE = "documentreference"
    ENCOUNTER = "encounter"
    NONE = None
    OBSERVATION = "observation"
    PATIENT = "patient"
    MEDICATION = "medication"
    MEDICATIONREQUEST = "medicationrequest"


@dataclass
class CountColumn:
    name: str
    db_type: str
    alias: str


def get_count_query(
    table_name: str,
    source_table: str,
    table_cols: list,
    min_subject: int = 10,
    where_clauses: list | None = None,
    fhir_resource: str | None = None,
    filter_resource: bool | None = True,
    patient_link: str = "subject_ref",
) -> str:
    """Generates count tables for generating study outputs"""
    path = Path(__file__).parent
    if fhir_resource not in {e.value for e in CountableFhirResource}:
        raise CountsBuilderError(
            f"Tried to create counts table for invalid resource {fhir_resource}."
        )

    table_col_classed = []
    for item in table_cols:
        # TODO: remove check after cutover
        if isinstance(item, list):
            table_col_classed.append(CountColumn(name=item[0], db_type=item[1], alias=item[2]))
        else:
            table_col_classed.append(CountColumn(name=item, db_type="varchar", alias=None))
    table_cols = table_col_classed

    query = base_templates.get_base_template(
        "count",
        path,
        table_name=table_name,
        source_table=source_table,
        table_cols=table_cols,
        min_subject=min_subject,
        where_clauses=where_clauses,
        fhir_resource=fhir_resource,
        filter_resource=filter_resource,
        patient_link=patient_link,
    )
    # workaround for conflicting sqlfluff enforcement
    return query.replace("-- noqa: disable=LT02\n", "")



import pathlib
import sys

import cumulus_fhir_support
import pyarrow

from cumulus_library import db_config, errors
from cumulus_library.databases import athena, base, duckdb


def _read_rows_for_resource(path: pathlib.Path, resource: str) -> list[dict]:
    rows = []

    # Support any ndjson files from the target folder directly
    rows += list(cumulus_fhir_support.read_multiline_json_from_dir(path, resource))

    # Also support being given an ETL output folder, and look in the table subdir
    subdir = path / resource.lower()
    rows += list(cumulus_fhir_support.read_multiline_json_from_dir(subdir, resource))

    return rows


def read_ndjson_dir(path: str) -> dict[str, pyarrow.Table]:
    """Loads a directory tree of raw ndjson into schema-ful tables.

    :param path: a directory path
    :returns: dictionary of table names (like 'documentreference') to table
      data (with schema)
    """
    all_tables = {}

    # Manually specify the list of resources because we want to create each table
    # even if the folder does not exist.
    resources = [
        "AllergyIntolerance",
        "Condition",
        "Device",
        "DiagnosticReport",
        "DocumentReference",
        "Encounter",
        "Immunization",
        "Medication",
        "MedicationRequest",
        "Observation",
        "Patient",
        "Procedure",
        "ServiceRequest",
    ]
    for resource in resources:
        table_name = resource.lower()
        rows = _read_rows_for_resource(pathlib.Path(path), resource)

        # Make a pyarrow table with full schema from the data
        schema = cumulus_fhir_support.pyarrow_schema_from_rows(resource, rows)
        all_tables[table_name] = pyarrow.Table.from_pylist(rows, schema)

    # And now some special support for a few ETL tables.
    metadata_tables = [
        "etl__completion",
        "etl__completion_encounters",
    ]
    for metadata_table in metadata_tables:
        rows = list(cumulus_fhir_support.read_multiline_json_from_dir(f"{path}/{metadata_table}"))
        if rows:
            # Auto-detecting the schema works for these simple tables
            all_tables[metadata_table] = pyarrow.Table.from_pylist(rows)

    return all_tables


def create_db_backend(args: dict[str, str]) -> (base.DatabaseBackend, str):
    """Retrieves a database backend and target schema from CLI args"""
    db_config.db_type = args["db_type"]

    if db_config.db_type == "duckdb":
        load_ndjson_dir = args.get("load_ndjson_dir")
        # TODO: reevaluate as DuckDB's local schema support evolves.
        # https://duckdb.org/docs/sql/statements/set.html#syntax
        if not (args.get("schema_name") is None or args["schema_name"] == "main"):
            print(
                "Warning - local schema names are not yet supported by duckDB's "
                "python library - using 'main' instead"
            )
        schema_name = "main"
        backend = duckdb.DuckDatabaseBackend(args["database"])
        if load_ndjson_dir:
            backend.insert_tables(read_ndjson_dir(load_ndjson_dir))
    elif db_config.db_type == "athena":
        if (
            args.get("schema_name") is not None
            and args.get("database") is not None
            and args.get("schema_name") != args.get("database")
        ):
            sys.exit(
                f"Two separate values, database: '{args['database']}' and "
                f"schema_name: '{args['schema_name']}' were supplied for "
                "an Athena database connection.\n"
                "Only one of these needs to be supplied. Both can be passed "
                "as arguments, but they must be equal"
            )
        schema_name = args["schema_name"] or args["database"]
        backend = athena.AthenaDatabaseBackend(
            args["region"],
            args["work_group"],
            args["profile"],
            schema_name,
        )
        if args.get("load_ndjson_dir"):
            sys.exit("Loading an ndjson dir is not supported with --db-type=athena.")
    else:
        raise errors.CumulusLibraryError(f"'{db_config.db_type}' is not a supported database.")

    return (backend, schema_name)



"""Holds enums used across more than one module"""

import enum


class ProtectedTableKeywords(enum.Enum):
    """Tables with a pattern like '_{keyword}_' are not manually dropped."""

    ETL = "etl"
    LIB = "lib"
    NLP = "nlp"


class ProtectedTables(enum.Enum):
    """Tables created by cumulus for persistence outside of study rebuilds"""

    STATISTICS = "lib_statistics"
    TRANSACTIONS = "lib_transactions"


class StatisticsTypes(enum.Enum):
    """A subset of workflows that create statistics sampling artifacts"""

    PSM = "psm"


class LogStatuses(enum.Enum):
    DEBUG = "debug"
    ERROR = "error"
    FINISHED = "finished"
    INFO = "info"
    RESUMED = "resumed"
    STARTED = "started"
    WARN = "warn"



"""Class for generating counts tables from templates"""

from pathlib import Path

from cumulus_library import BaseTableBuilder, errors, study_manifest
from cumulus_library.builders.statistics_templates import counts_templates

# Defined here for easy overriding by tests
DEFAULT_MIN_SUBJECT = 10


class CountsBuilder(BaseTableBuilder):
    """Extends BaseTableBuilder for counts-related use cases"""

    def __init__(self, study_prefix: str | None = None):
        super().__init__()
        self.study_prefix = study_prefix

    def get_table_name(self, table_name: str, duration=None) -> str:
        """Convenience method for constructing table name

        :param table_name: table name to add after the study prefix
        :param duration: a time period reflecting the table binning strategy
        """
        if not self.study_prefix:
            raise errors.CountsBuilderError(
                "CountsBuilder must be either initiated with a study prefix, "
                "or be in a directory with a valid manifest.toml"
            )
        if duration:
            return f"{self.study_prefix}__{table_name}_{duration}"
        else:
            return f"{self.study_prefix}__{table_name}"

    def get_where_clauses(
        self, clause: list | str | None = None, min_subject: int | None = None
    ) -> list[str]:
        """Convenience method for constructing arbitrary where clauses.

        :param clause: either a string or a list of sql where statements
        :param min_subject: if clause is none, the bin size for a cnt_subject_ref filter
            (deprecated, use count_[fhir_resource](min_subject) instead)
        """
        if min_subject is None:
            min_subject = DEFAULT_MIN_SUBJECT
        if clause is None:
            return [f"cnt_subject_ref >= {min_subject}"]
        elif isinstance(clause, str):
            return [clause]
        elif isinstance(clause, list):
            return clause
        else:
            raise errors.CountsBuilderError(f"get_where_clauses invalid clause {clause}")

    def get_count_query(
        self, table_name: str, source_table: str, table_cols: list, **kwargs
    ) -> str:
        """Generates a counts table using a template

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :keyword where_clauses: An array of where clauses to use for filtering the data
        :keyword min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        :keyword fhir_resource: The type of FHIR resource to count (see
            statistics/statistics_templates/count_templates.CountableFhirResource)
        """
        if not table_name or not source_table or not table_cols:
            raise errors.CountsBuilderError(
                "count_query missing required arguments. " f"output table: {table_name}"
            )
        for key in kwargs:
            if key not in [
                "min_subject",
                "where_clauses",
                "fhir_resource",
                "filter_resource",
                "patient_link",
            ]:
                raise errors.CountsBuilderError(f"count_query received unexpected key: {key}")
        if "min_subject" in kwargs and kwargs["min_subject"] is None:
            kwargs["min_subject"] = DEFAULT_MIN_SUBJECT
        return counts_templates.get_count_query(table_name, source_table, table_cols, **kwargs)

    # ----------------------------------------------------------------------
    # The following function all wrap get_count_query as convenience methods.
    # We're not trying to be overly clever about this to persist the docstrings as the
    # primary interface that study authors would see when using these functions.

    def count_allergyintolerance(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
    ) -> str:
        """wrapper method for constructing allergyintolerance counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :param where_clauses: An array of where clauses to use for filtering the data
        :param min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            fhir_resource="allergyintolerance",
            patient_link="patient_ref",
        )

    def count_condition(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
    ) -> str:
        """wrapper method for constructing condition counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :param where_clauses: An array of where clauses to use for filtering the data
        :param min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            fhir_resource="condition",
            filter_resource=True,
        )

    def count_diagnosticreport(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
    ) -> str:
        """wrapper method for constructing diagnosticreport counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :param where_clauses: An array of where clauses to use for filtering the data
        :param min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            fhir_resource="diagnosticreport",
        )

    def count_documentreference(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
    ) -> str:
        """wrapper method for constructing documentreference counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :param where_clauses: An array of where clauses to use for filtering the data
        :param min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            fhir_resource="documentreference",
            filter_resource=True,
        )

    def count_encounter(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
    ) -> str:
        """wrapper method for constructing encounter counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :param where_clauses: An array of where clauses to use for filtering the data
        :param min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            fhir_resource="encounter",
        )

    def count_medicationrequest(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
    ) -> str:
        """wrapper method for constructing medicationrequests counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :param where_clauses: An array of where clauses to use for filtering the data
        :param min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            fhir_resource="medicationrequest",
        )

    def count_observation(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
    ) -> str:
        """wrapper method for constructing observation counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :param where_clauses: An array of where clauses to use for filtering the data
        :param min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            fhir_resource="observation",
        )

    def count_patient(
        self,
        table_name: str,
        source_table: str,
        table_cols: list,
        where_clauses: list | None = None,
        min_subject: int | None = None,
    ) -> str:
        """wrapper method for constructing patient counts tables

        :param table_name: The name of the table to create. Must start with study prefix
        :param source_table: The table to create counts data from
        :param table_cols: The columns from the source table to add to the count table
        :param where_clauses: An array of where clauses to use for filtering the data
        :param min_subject: An integer setting the minimum bin size for inclusion
            (default: 10)
        """
        return self.get_count_query(
            table_name,
            source_table,
            table_cols,
            where_clauses=where_clauses,
            min_subject=min_subject,
            fhir_resource="patient",
        )

    # End of wrapper section
    # ----------------------------------------------------------------------

    def write_counts(self, filepath: str):
        """Convenience method for writing counts queries to disk

        :param filepath: path to file to write queries out to.
        """
        self.prepare_queries()
        self.comment_queries()
        self.write_queries(path=Path(filepath))

    def prepare_queries(
        self, *args, manifest: study_manifest.StudyManifest | None = None, **kwargs
    ):
        """Prepare count queries

        This should be overridden in any count generator. See studies/core/count_core.py
        for an example
        """
        if manifest and not self.study_prefix:
            self.study_prefix = manifest.get_study_prefix()



import sys

from rich.progress import Progress, TaskID

from cumulus_library import base_utils, databases, enums, errors, study_manifest
from cumulus_library.template_sql import base_templates


def _execute_drop_queries(
    cursor: databases.DatabaseCursor,
    verbose: bool,
    view_table_list: list,
    progress: Progress,
    task: TaskID,
) -> None:
    """Handler for executing drop view/table queries and displaying console output.

    :param cursor: A DatabaseCursor object
    :param verbose: toggle from progress bar to query output
    :param view_table_list: a list of views and tables beginning with
      the study prefix
    :param progress: a rich progress bar renderer
    :param task: a TaskID for a given progress bar
    """
    for view_table in view_table_list:
        drop_view_table = base_templates.get_drop_view_table(
            name=view_table[0], view_or_table=view_table[1]
        )
        with base_utils.query_console_output(verbose, drop_view_table, progress, task):
            cursor.execute(drop_view_table)


def _get_unprotected_stats_view_table(
    config: base_utils.StudyConfig,
    query: str,
    artifact_type: str,
    drop_prefix: str,
    display_prefix: str,
    stats_clean: bool,
):
    """Gets all items from the database by type, less any protected items

    :param config: A StudyConfig object
    :param query: A query to get the raw list of items from the db
    :param artifact_type: either 'table' or 'view'
    :param drop_prefix: The prefix requested to drop
    :param display_prefix: The expected study prefix
    :param stats_clean: A boolean indicating if stats tables are being cleaned

    :returns: a list of study tables to drop
    """
    cursor = config.db.cursor()
    unprotected_list = []
    db_contents = cursor.execute(query).fetchall()
    if (
        f"{drop_prefix}{enums.ProtectedTables.STATISTICS.value}",
    ) in db_contents and not stats_clean:
        protected_list = cursor.execute(
            f"""SELECT {artifact_type.lower()}_name 
            FROM {drop_prefix}{enums.ProtectedTables.STATISTICS.value}
            WHERE study_name = '{display_prefix}'"""  # noqa: S608
        ).fetchall()
        for protected_tuple in protected_list:
            if protected_tuple in db_contents:
                db_contents.remove(protected_tuple)
    for db_row_tuple in db_contents:
        # this check handles athena reporting views as also being tables,
        # so we don't waste time dropping things that don't exist
        if artifact_type == "TABLE":
            if not any(db_row_tuple[0] in iter_q_and_t for iter_q_and_t in unprotected_list):
                unprotected_list.append([db_row_tuple[0], artifact_type])
        else:
            unprotected_list.append([db_row_tuple[0], artifact_type])
    return unprotected_list


def clean_study(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest | None,
    prefix: str | None = None,
) -> list:
    """Removes tables beginning with the study prefix from the database schema

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword prefix: override manifest-based prefix discovery with the provided prefix
    :returns: list of dropped tables (for unit testing only)

    """
    if not prefix and not manifest:
        raise errors.CumulusLibraryError(
            "Either a manifest parser or a filter prefix must be provided"
        )
    if manifest and manifest.get_dedicated_schema():
        drop_prefix = ""
        display_prefix = ""
    elif not prefix:
        drop_prefix = f"{manifest.get_study_prefix()}__"
        display_prefix = manifest.get_study_prefix()
    else:
        drop_prefix = prefix
        display_prefix = drop_prefix

    if config.stats_clean:
        confirm = input(
            "This will remove all historical stats tables in the "
            f"{display_prefix} study - are you sure? (y/N)"
        )
        if confirm is None or confirm.lower() not in ("y", "yes"):
            sys.exit("Table cleaning aborted")

    cursor = config.db.cursor()
    view_sql = base_templates.get_show_views(config.schema, drop_prefix)
    table_sql = base_templates.get_show_tables(config.schema, drop_prefix)
    view_table_list = []
    for query_and_type in [[view_sql, "VIEW"], [table_sql, "TABLE"]]:
        view_table_list += _get_unprotected_stats_view_table(
            config,
            query_and_type[0],
            query_and_type[1],
            drop_prefix,
            display_prefix,
            config.stats_clean,
        )
    if not view_table_list:
        return view_table_list

    # We'll do a pass to see if any of these tables were created outside of a
    # study builder, and remove them from the list.
    for view_table in view_table_list.copy():
        if any(
            (
                f"__{word.value}_" in view_table[0]
                or view_table[0].endswith(f"__{word.value}")
                or view_table[0].startswith(f"{word.value}_")
            )
            for word in enums.ProtectedTableKeywords
        ):
            view_table_list.remove(view_table)
    if prefix:
        print("The following views/tables were selected by prefix:")
        for view_table in view_table_list:
            print(f"  {view_table[0]}")
        confirm = input("Remove these tables? (y/N)")
        if confirm.lower() not in ("y", "yes"):
            sys.exit("Table cleaning aborted")
    if dedicated := manifest.get_dedicated_schema():
        view_table_list = [
            (
                f"`{dedicated}`.`{x[0]}`",
                x[1],
            )
            for x in view_table_list
        ]
    # We want to only show a progress bar if we are :not: printing SQL lines
    with base_utils.get_progress_bar(disable=config.verbose) as progress:
        task = progress.add_task(
            f"Removing {display_prefix} study artifacts...",
            total=len(view_table_list),
            visible=not config.verbose,
        )
        _execute_drop_queries(
            cursor,
            config.verbose,
            view_table_list,
            progress,
            task,
        )
    # if we're doing a stats clean, we'll also remove the table containing the
    # list of protected tables
    if config.stats_clean:
        drop_query = base_templates.get_drop_view_table(
            f"{drop_prefix}{enums.ProtectedTables.STATISTICS.value}", "TABLE"
        )
        cursor.execute(drop_query)

    return view_table_list



"""Handles the creation of new tables"""

import contextlib
import importlib.util
import inspect
import pathlib
import re
import sys
import tomllib
import zipfile

from rich.progress import Progress, TaskID

from cumulus_library import (
    BaseTableBuilder,
    CountsBuilder,
    base_utils,
    databases,
    enums,
    errors,
    log_utils,
    study_manifest,
)
from cumulus_library.builders import protected_table_builder, psm_builder, valueset_builder


@contextlib.contextmanager
def _temporary_sys_path(add: pathlib.Path) -> None:
    orig_path = list(sys.path)
    try:
        sys.path.insert(0, str(add))
        yield
    finally:
        sys.path = orig_path


def _load_and_execute_builder(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    filename: str,
    db_parser: databases.DatabaseParser = None,
    write_reference_sql: bool = False,
    doc_str: str | None = None,
    prepare: bool = False,
    data_path: pathlib.Path,
    query_count: int | None = None,
) -> int:
    """Loads a table builder from a file.

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword filename: filename of a module implementing a TableBuilder
    :keyword db_parser: an object implementing DatabaseParser for the target database
    :keyword write_reference_sql: if true, writes sql to disk inside a study's directory
    :keyword doc_str: A string to insert between queries written to disk
    :keyword prepare: If true, will render query instead of executing
    :keyword data_path: If prepare is true, the path to write rendered data to
    :keyword query_count: if prepare is true, the number of queries already rendered
    :returns: number of processed queries
    """

    # Since we have to support arbitrary user-defined python files here, we
    # jump through some importlib hoops to import the module directly from
    # a source file defined in the manifest.
    spec = importlib.util.spec_from_file_location(
        "table_builder", f"{manifest._study_path}/{filename}"
    )
    table_builder_module = importlib.util.module_from_spec(spec)
    sys.modules["table_builder"] = table_builder_module
    # Inject the study dir into sys.path so that builders can import
    # from surrounding utility code, even if the study isn't installed.
    # (i.e. you're working from a git checkout and do something like `-s .`)
    with _temporary_sys_path(manifest._study_path.parent):
        spec.loader.exec_module(table_builder_module)

    # We're going to find all subclasses of BaseTableBuild in this file.
    # Since BaseTableBuilder itself is a valid subclass of BaseTableBuilder,
    # we'll detect and skip it. If we don't find any subclasses,
    # we'll punt.
    table_builder_subclasses = []
    for _, cls_obj in inspect.getmembers(table_builder_module, inspect.isclass):
        if issubclass(cls_obj, BaseTableBuilder) and cls_obj != BaseTableBuilder:
            table_builder_subclasses.append(cls_obj)

    if len(table_builder_subclasses) == 0:
        raise errors.StudyManifestParsingError(
            f"Error loading {manifest._study_path}{filename}\n"
            "Custom builders must extend the BaseTableBuilder class."
        )
    # Remove instances of intermediate classes, if present (usually not)
    table_builder_subclasses = list(
        filter(lambda x: x.__name__ != "CountsBuilder", table_builder_subclasses)
    )

    # We'll get the subclass, initialize it, run the executor code, and then
    # remove it so it doesn't interfere with the next python module to
    # execute, since the subclass would otherwise hang around.
    table_builder_class = table_builder_subclasses[0]
    if issubclass(table_builder_class, CountsBuilder):
        table_builder = table_builder_class(study_prefix=manifest.get_study_prefix())
    else:
        table_builder = table_builder_class()
    if write_reference_sql:
        prefix = manifest.get_study_prefix()
        table_builder.prepare_queries(config=config, manifest=manifest, parser=db_parser)
        for query_pos in range(len(table_builder.queries)):
            if "s3://" in table_builder.queries[query_pos]:
                table_builder.queries[query_pos] = re.sub(
                    f"s3://(.+){prefix}",
                    f"s3://bucket/db_path/{prefix}",
                    table_builder.queries[query_pos],
                )
        table_builder.comment_queries(doc_str=doc_str)
        new_filename = pathlib.Path(f"{filename}").stem + ".sql"
        table_builder.write_queries(
            path=pathlib.Path(f"{manifest._study_path}/reference_sql/" + new_filename)
        )
    elif prepare:
        table_builder.prepare_queries(
            config=config,
            manifest=manifest,
            parser=db_parser,
        )
        _render_output(
            manifest.get_study_prefix(), table_builder.queries, data_path, filename, query_count
        )
    else:
        table_builder.execute_queries(
            config=config,
            manifest=manifest,
            parser=db_parser,
        )
    num_queries = len(table_builder.queries)
    # After running the executor code, we'll remove
    # it so it doesn't interfere with the next python module to
    # execute, since the subclass would otherwise hang around.
    del sys.modules[table_builder_module.__name__]
    del table_builder_module

    return num_queries


def run_protected_table_builder(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
) -> None:
    """Creates protected tables for persisting selected data across runs

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    """
    ptb = protected_table_builder.ProtectedTableBuilder()
    ptb.execute_queries(
        config=config,
        manifest=manifest,
    )


def _run_workflow(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    filename: str,
    prepare: str,
    data_path: pathlib.Path,
    query_count: int,
) -> int:
    """Loads workflow config from toml definitions and executes workflow

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :param filename: Filename of the workflow config
    :param prepare: If true, will render query instead of executing
    :param data_path: If prepare is true, the path to write rendered data to
    :param query_count: if prepare is true, the number of queries already rendered
    :returns: a count of processed queries
    """
    toml_path = pathlib.Path(f"{manifest._study_path}/{filename}")
    if prepare:
        with open(toml_path, encoding="utf-8") as file:
            workflow_config = file.read()
        _render_output(
            manifest.get_study_prefix(),
            [workflow_config],
            data_path,
            filename,
            query_count,
            is_toml=True,
        )
        return 1
    existing_stats = []
    if not config.stats_build:
        existing_stats = (
            config.db.cursor()
            .execute(
                "SELECT view_name FROM "  # noqa: S608
                f"{manifest.get_study_prefix()}__{enums.ProtectedTables.STATISTICS.value}"
            )
            .fetchall()
        )
    # This open is a bit redundant with the open inside of the PSM builder,
    # but we're letting it slide so that builders function similarly
    # across the board
    safe_timestamp = base_utils.get_tablename_safe_iso_timestamp()
    with open(toml_path, "rb") as file:
        workflow_config = tomllib.load(file)
        config_type = workflow_config["config_type"]
        target_table = workflow_config.get("target_table", workflow_config.get("table_prefix", ""))

    if (target_table,) in existing_stats and not config.stats_build:
        return 0
    match config_type:
        case "psm":
            builder = psm_builder.PsmBuilder(
                toml_config_path=toml_path,
                config=workflow_config,
                data_path=manifest.data_path / f"{manifest.get_study_prefix()}/psm",
            )
        case "valueset":
            builder = valueset_builder.ValuesetBuilder(
                toml_config_path=toml_path,
                config=workflow_config,
                data_path=manifest.data_path / f"{manifest.get_study_prefix()}/valueset",
            )
        case _:  # pragma: no cover
            raise errors.StudyManifestParsingError(
                f"{toml_path} references an invalid workflow type {config_type}."
            )
    builder.execute_queries(
        config=config,
        manifest=manifest,
        table_suffix=safe_timestamp,
    )
    if config_type in set(item.value for item in enums.StatisticsTypes):
        log_utils.log_statistics(
            config=config,
            manifest=manifest,
            table_type=config_type,
            table_name=f"{target_table}_{safe_timestamp}",
            view_name=target_table,
        )
    return len(builder.queries)


def build_matching_files(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    builder: str | None,
    db_parser: databases.DatabaseParser = None,
    prepare: bool,
    data_path: pathlib.Path,
):
    """targets all table builders matching a target string for running

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword builder: filename of a module implementing a TableBuilder
    :keyword db_parser: an object implementing DatabaseParser for the target database
    :keyword prepare: If true, will render query instead of executing
    :keyword data_path: If prepare is true, the path to write rendered data to
    """
    if prepare:
        _check_if_preparable(manifest.get_study_prefix())  # pragma: no cover
    all_generators = manifest.get_all_generators()
    matches = []
    if not builder:  # pragma: no cover
        matches = all_generators
    else:
        for file in all_generators:
            if file.find(builder) != -1:
                matches.append(file)
    build_study(
        config,
        manifest,
        db_parser=db_parser,
        file_list=matches,
        prepare=prepare,
        data_path=data_path,
    )


def build_study(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    db_parser: databases.DatabaseParser = None,
    continue_from: str | None = None,
    file_list: list | None = None,
    prepare: bool,
    data_path: pathlib.Path | None,
) -> None:
    """Creates tables in the schema by iterating through the sql_config.file_names

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword db_parser: a parser for the target database
    :keyword continue_from: Name of a file to resume table creation from
    :keyword prepare: If true, will render query instead of executing
    :keyword data_path: If prepare is true, the path to write rendered data to
    """
    if prepare:
        _check_if_preparable(manifest.get_study_prefix())
    if file_list is None:
        file_list = manifest.get_file_list(continue_from)
    if prepare:
        data_dir = data_path / manifest.get_study_prefix()
        for file in data_dir.glob("*"):
            if file.is_file():  # pragma: no cover
                file.unlink()
    query_count = 0
    for file in file_list:
        if file.endswith(".py"):
            query_count += _load_and_execute_builder(
                config=config,
                manifest=manifest,
                filename=file,
                db_parser=db_parser,
                data_path=data_path,
                prepare=prepare,
                query_count=query_count,
            )
        elif file.endswith(".toml"):
            query_count += _run_workflow(
                config=config,
                manifest=manifest,
                filename=file,
                data_path=data_path,
                prepare=prepare,
                query_count=query_count,
            )
        elif file.endswith(".sql"):
            query_count += _run_raw_queries(
                config=config,
                manifest=manifest,
                filename=file,
                data_path=data_path,
                prepare=prepare,
                query_count=query_count,
            )
        else:
            raise errors.StudyManifestParsingError
    if prepare:
        with zipfile.ZipFile(
            f"{data_path}/{manifest.get_study_prefix()}.zip", "w", zipfile.ZIP_DEFLATED
        ) as z:
            for file in data_dir.iterdir():
                z.write(file, file.relative_to(data_dir))


def _run_raw_queries(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    filename: str,
    *,
    data_path: pathlib.Path | None,
    prepare: bool,
    query_count: int,
) -> int:
    """Creates tables in the schema by iterating through the sql_config.file_names

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :param filename: the name of the sql file to read
    :param prepare: If true, will render query instead of executing
    :param data_path: If prepare is true, the path to write rendered data to
    :param query_count: the number of queries currently processed
    :returns number of queries processed:
    """
    queries = []
    for query in base_utils.parse_sql(base_utils.load_text(f"{manifest._study_path}/{filename}")):
        queries.append([query, filename])
    for query in queries:
        query[0] = base_utils.update_query_if_schema_specified(query[0], manifest)
        query[0] = query[0].replace(
            f"`{manifest.get_study_prefix()}__",
            "`",
        )
    if prepare:
        _render_output(
            manifest.get_study_prefix(), [q[0] for q in queries], data_path, filename, query_count
        )
    else:
        # We'll explicitly create a cursor since recreating cursors for each
        # table in a study is slightly slower for some databases
        cursor = config.db.cursor()
        # We want to only show a progress bar if we are :not: printing SQL lines
        with base_utils.get_progress_bar(disable=config.verbose) as progress:
            task = progress.add_task(
                f"Building tables from {filename}...",
                total=len(queries),
                visible=not config.verbose,
            )
            _execute_build_queries(
                config=config,
                manifest=manifest,
                cursor=cursor,
                queries=queries,
                progress=progress,
                task=task,
            )
    return len(queries)


def _render_output(
    study_name: str,
    outputs: list,
    data_path: pathlib.Path,
    filename: str,
    count: int,
    *,
    is_toml: bool = False,
):
    if is_toml:
        suffix = "toml"
    else:
        suffix = "sql"
    for index, output in enumerate(outputs):
        if is_toml:
            name = "config"
        else:
            # This regex attempts to discover the table name, via looking for the first
            # dunder, and then gets the start of the line its on as a non-quote requiring
            # part of a file name. So for example, finding SQL that looks like this:
            #   CREATE TABLE foo__bar AS (varchar baz)
            # would result in `create_table_foo__bar` being assigned to name. The goal
            # is to make this at least mildly parsable at the file system level if someone
            # is reviewing a prepared study
            name = re.search(r"(.*)__\w*", output)[0].lower().replace(" ", "_")
        total_index = count + index
        new_filename = f"{total_index:04d}.{filename.rsplit('.', 1)[0]}.{index:02d}.{name}.{suffix}"
        file_path = data_path / f"{study_name}/{new_filename}"

        file_path.parent.mkdir(exist_ok=True, parents=True)
        with open(file_path, "w", encoding="UTF-8") as f:
            f.write(output)


def _check_if_preparable(prefix):
    # This list should include any study which requires interrogating the database to
    # find if data is available to query (outside of a toml-driven workflow),
    # which isn't doable as a distributed query
    if prefix in ["core", "discovery", "data-metrics"]:
        sys.exit(
            f"Study '{prefix}'' does not support prepare mode. It must be run "
            "directly against a target database."
        )


def _query_error(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    query_and_filename: list,
    exit_message: str,
) -> None:
    print(
        "An error occured executing the following query in ",
        f"{query_and_filename[1]}:",
        file=sys.stderr,
    )
    print("--------", file=sys.stderr)
    print(query_and_filename[0], file=sys.stderr)
    print("--------", file=sys.stderr)
    log_utils.log_transaction(config=config, manifest=manifest, status=enums.LogStatuses.ERROR)
    sys.exit(exit_message)


def _execute_build_queries(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    cursor: databases.DatabaseCursor,
    queries: list,
    progress: Progress,
    task: TaskID,
) -> None:
    """Handler for executing create table queries and displaying console output.

    :param manifest: a StudyManifest object
    :param cursor: A DatabaseCursor object
    :param queries: a list of queries read from files in sql_config.file_names
    :param progress: a rich progress bar renderer
    :param task: a TaskID for a given progress bar
    """
    for query in queries:
        create_line = query[0].split("\n")[0]
        if (
            f" {manifest.get_study_prefix()}__" not in create_line
            and not manifest.get_dedicated_schema()
        ):
            _query_error(
                config,
                manifest,
                query,
                "This query does not contain the study prefix. All tables should "
                f"start with a string like `{manifest.get_study_prefix()}__`, "
                "and it should be in the first line of the query.",
            )
        if any(
            f" {manifest.get_study_prefix()}__{word.value}_" in create_line
            for word in enums.ProtectedTableKeywords
        ):
            _query_error(
                config,
                manifest,
                query,
                "This query contains a table name which contains a reserved word "
                "immediately after the study prefix. Please rename this table so "
                "that is does not begin with one of these special words "
                "immediately after the double undescore.\n Reserved words: "
                f"{(word.value for word in enums.ProtectedTableKeywords)}",
            )
        if create_line.count("__") > 1:
            _query_error(
                config,
                manifest,
                query,
                "This query contains a table name with more than one '__' in it. "
                "Double underscores are reserved for special use cases. Please "
                "rename this table so the only double undercore is after the "
                f"study prefix, e.g. `{manifest.get_study_prefix()}__`",
            )
        try:
            with base_utils.query_console_output(config.verbose, query[0], progress, task):
                cursor.execute(query[0])
        except Exception as e:  # pylint: disable=broad-exception-caught
            _query_error(
                config,
                manifest,
                query,
                "You can debug issues with this query using `sqlfluff lint`, "
                "or by executing the query directly against the database.\n"
                f"Error: {e}",
            )



import pathlib

import pyarrow
from pyarrow import csv, parquet
from rich.progress import track

from cumulus_library import base_utils, study_manifest
from cumulus_library.template_sql import base_templates

# Database exporting functions


def reset_counts_exports(
    manifest: study_manifest.StudyManifest,
) -> None:
    """
    Removes exports associated with this study from the ../data_export directory.
    """
    path = pathlib.Path(f"{manifest.data_path}/{manifest.get_study_prefix()}")
    if path.exists():
        # we're just going to remove the count exports - stats exports in
        # subdirectories are left alone by this call
        for file in path.glob("*.*"):
            file.unlink()


def _write_chunk(writer, chunk, arrow_schema):
    writer.write(
        pyarrow.Table.from_pandas(
            chunk.sort_values(by=list(chunk.columns), ascending=False, na_position="first"),
            preserve_index=False,
            schema=arrow_schema,
        )
    )


def export_study(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    *,
    data_path: pathlib.Path,
    archive: bool,
    chunksize: int = 1000000,
) -> list:
    """Exports csvs/parquet extracts of tables listed in export_list
    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword data_path: the path to the place on disk to save data
    :keyword archive: If true, get all study data and zip with timestamp
    :keyword chunksize: number of rows to export in a single transaction
    :returns: a list of queries, (only for unit tests)
    """
    reset_counts_exports(manifest)
    if manifest.get_dedicated_schema():
        prefix = f"{manifest.get_dedicated_schema()}."
    else:
        prefix = f"{manifest.get_study_prefix()}__"
    if archive:
        table_query = base_templates.get_show_tables(config.schema, prefix)
        result = config.db.cursor().execute(table_query).fetchall()
        table_list = manifest.get_export_table_list()
        for row in result:
            if row[0] not in table_list:
                table_list.append(study_manifest.ManifestExport(name=row[0], export_type="archive"))
    else:
        table_list = manifest.get_export_table_list()
    queries = []
    path = pathlib.Path(f"{data_path}/{manifest.get_study_prefix()}/")
    for table in track(
        table_list,
        description=f"Exporting {manifest.get_study_prefix()} data...",
    ):
        query = f"SELECT * FROM {table.name}"  # noqa: S608
        query = base_utils.update_query_if_schema_specified(query, manifest)
        dataframe_chunks, db_schema = config.db.execute_as_pandas(query, chunksize=chunksize)
        path.mkdir(parents=True, exist_ok=True)
        arrow_schema = pyarrow.schema(config.db.col_pyarrow_types_from_sql(db_schema))
        with parquet.ParquetWriter(
            f"{path}/{table.name}.{table.export_type}.parquet", arrow_schema
        ) as p_writer:
            with csv.CSVWriter(
                f"{path}/{table.name}.{table.export_type}.csv",
                arrow_schema,
                write_options=csv.WriteOptions(
                    # Note that this quoting style is not exactly csv.QUOTE_MINIMAL
                    # https://github.com/apache/arrow/issues/42032
                    quoting_style="needed"
                ),
            ) as c_writer:
                for chunk in dataframe_chunks:
                    _write_chunk(p_writer, chunk, arrow_schema)  # pragma: no cover
                    _write_chunk(c_writer, chunk, arrow_schema)  # pragma: no cover
        queries.append(query)
    if archive:
        base_utils.zip_dir(path, data_path, manifest.get_study_prefix())
    return queries



import pathlib

from cumulus_library.template_sql import base_templates


def get_system_pairs(output_table_name: str, code_system_tables: list) -> str:
    """Extracts code system details as a standalone table"""

    # Since it's easier to wrangle data before SQL, this code block does
    # the following: given a datastructure like:
    #     [('a',list),('b', dict),('c', dict),('d', list), ('e', dict)]
    # Since data access in sql by nested dicts is just joining operators,
    # it will flatten columns together by combining dicts up to the next
    # list instance, i.e. when the next unnest would be needed:
    #     [('a',list),('b.c.d', list), ('e', dict)]
    # It also creates a display name along the way
    for table in code_system_tables:
        unnest_layer = ""
        squashed_hierarchy = []
        display_col = ""
        for column in table["column_hierarchy"]:
            unnest_layer = ".".join(x for x in [unnest_layer, column[0]] if x)
            display_col = ".".join(x for x in [display_col, column[0]] if x)
            if column[1] is list:
                squashed_hierarchy.append((unnest_layer, list))
                unnest_layer = ""
        if unnest_layer != "":
            squashed_hierarchy.append((unnest_layer, dict))
        table["column_hierarchy"] = squashed_hierarchy
        table["column_display_name"] = display_col.removesuffix(".coding")

    return base_templates.get_base_template(
        "code_system_pairs",
        pathlib.Path(__file__).parent,
        output_table_name=output_table_name,
        code_system_tables=code_system_tables,
    )



import cumulus_library
from cumulus_library.studies.core.core_templates import core_templates
from cumulus_library.template_sql import sql_utils

expected_table_cols = {
    "allergyintolerance": {
        "id": [],
        "type": [],
        "category": [],
        "criticality": [],
        "recordedDate": [],
        "patient": sql_utils.REFERENCE,
        "encounter": sql_utils.REFERENCE,
        "reaction": ["severity"],
    }
}


class CoreAllergyIntoleranceBuilder(cumulus_library.BaseTableBuilder):
    display_text = "Creating AllergyIntolerance tables..."

    def prepare_queries(self, *args, config: cumulus_library.StudyConfig, **kwargs):
        code_sources = [
            sql_utils.CodeableConceptConfig(
                source_table="allergyintolerance",
                column_hierarchy=[("clinicalStatus", dict)],
                target_table="core__allergyintolerance_dn_clinical_status",
                filter_priority=True,
                code_systems=[
                    # Restrict to just this required binding system
                    "http://terminology.hl7.org/CodeSystem/allergyintolerance-clinical",
                ],
            ),
            sql_utils.CodeableConceptConfig(
                source_table="allergyintolerance",
                column_hierarchy=[("verificationStatus", dict)],
                target_table="core__allergyintolerance_dn_verification_status",
                filter_priority=True,
                code_systems=[
                    # Restrict to just this required binding system
                    "http://terminology.hl7.org/CodeSystem/allergyintolerance-verification",
                ],
            ),
            sql_utils.CodeableConceptConfig(
                source_table="allergyintolerance",
                column_hierarchy=[("code", dict)],
                target_table="core__allergyintolerance_dn_code",
            ),
            sql_utils.CodeableConceptConfig(
                source_table="allergyintolerance",
                column_hierarchy=[("reaction", list), ("substance", dict)],
                target_table="core__allergyintolerance_dn_reaction_substance",
                expected={"substance": sql_utils.CODEABLE_CONCEPT},
            ),
            sql_utils.CodeableConceptConfig(
                source_table="allergyintolerance",
                column_hierarchy=[("reaction", list), ("manifestation", list)],
                target_table="core__allergyintolerance_dn_reaction_manifestation",
                expected={"manifestation": sql_utils.CODEABLE_CONCEPT},
            ),
        ]
        self.queries += sql_utils.denormalize_complex_objects(config.db, code_sources)
        validated_schema = sql_utils.validate_schema(config.db, expected_table_cols)
        self.queries.append(
            core_templates.get_core_template("allergyintolerance", validated_schema)
        )



"""Collection of jinja template getters for common SQL queries"""

import enum
import pathlib

import jinja2
import pandas

from cumulus_library import db_config, errors
from cumulus_library.template_sql import sql_utils


class TableView(enum.Enum):
    """Convenience enum for building drop queries"""

    TABLE = "TABLE"
    VIEW = "VIEW"


def get_base_template(filename_stem: str, path: pathlib.Path | None = None, **kwargs: dict) -> str:
    """Abstract renderer for jinja templates

    You can use this renderer directly, but if you are designing a function
    that you expect to be commonly used, it's recommended to create a context
    specific loader function instead.

    This function will autoload macros in cumulus_library/template_sql/shared_macros,
    as well as any macros in a folder provided by path
    """
    base_path = pathlib.Path(__file__).parent
    with open(f"{path or base_path}/{filename_stem}.sql.jinja") as file:
        template = file.read()
        macro_paths = [base_path / "shared_macros"]
        if path:
            macro_paths.append(path)
        loader = jinja2.FileSystemLoader(macro_paths)
        env = jinja2.Environment(loader=loader).from_string(template)  # noqa: S701
        return env.render(**kwargs, db_type=db_config.db_type)


# All remaining functions are context-specific calls aimed at providing
# guidance around table creation for commonly used SQL functions


def get_alias_table_query(source_table: str, target_table: str):
    """Creates a view of source_table named target_table"""
    return get_base_template("alias_table", source_table=source_table, target_table=target_table)


def get_codeable_concept_denormalize_query(
    config: sql_utils.CodeableConceptConfig,
) -> str:
    """extracts codeable concepts from a specified table.

    This function is targeted at arbitrary codeableConcept elements - see
    http://hl7.org/fhir/datatypes-definitions.html#CodeableConcept for more info.
    This may be or may not be an array field depending on the context of use -
    check the specification of the specific resource you're interested in.
    See the CodeableConceptConfig for details on how to handle array vs non-
    array use cases.

    :param config: a CodableConeptConfig
    """

    if len(config.column_hierarchy) > 2:
        raise errors.CumulusLibraryError(
            "base_templates.get_codeable_concept_denormalize_query expects a column "
            f"hierarchy of length 2 or less, but received: {config.column_hierarchy}"
        )
    # If we get a None for code systems, we want one dummy value so the jinja
    # for loop will do a single pass. This implicitly means that we're not
    # filtering, so this parameter will be otherwise ignored
    config.code_systems = config.code_systems or ["all"]
    return get_base_template(
        "codeable_concept_denormalize",
        source_table=config.source_table,
        source_id=config.source_id,
        column_name=config.column_hierarchy[-1][0],
        parent_field=None if len(config.column_hierarchy) == 1 else config.column_hierarchy[0][0],
        is_array=(config.column_hierarchy[0][1] is list),
        child_is_array=False
        if len(config.column_hierarchy) == 1
        else config.column_hierarchy[1][1] is list,
        target_table=config.target_table,
        filter_priority=config.filter_priority,
        code_systems=config.code_systems,
        extra_fields=config.extra_fields or [],
    )


def get_coding_denormalize_query(
    config: sql_utils.CodingConfig,
) -> str:
    """extracts codings from a specified table.

    This function reimplements get_codeable_concept_denormalize_query targeted
    at a bare coding element

    TODO: this is temporary and this should be replaced by a generic DN
    query.

    :param config: a CodingConfig
    """
    if len(config.column_hierarchy) != 2:
        raise errors.CumulusLibraryError(
            "base_templates.get_coding_denormalize_query expects a column "
            f"hierarchy of length 2, but received: {config.column_hierarchy}"
        )
    # If we get a None for code systems, we want one dummy value so the jinja
    # for loop will do a single pass. This implicitly means that we're not
    # filtering, so this parameter will be otherwise ignored
    config.code_systems = config.code_systems or ["all"]
    return get_base_template(
        "coding_denormalize",
        source_table=config.source_table,
        source_id=config.source_id,
        column_name=config.column_hierarchy[1][0],
        parent_field=config.column_hierarchy[0][0],
        target_table=config.target_table,
        filter_priority=config.filter_priority,
        code_systems=config.code_systems,
    )


def get_column_datatype_query(
    schema_name: str,
    table_names: str | list,
    column_names: list | None = None,
    include_table_names: bool | None = False,
):
    """Gets the in-database data representation of a given column"""
    if isinstance(table_names, str):
        table_names = [table_names]
    return get_base_template(
        "column_datatype",
        schema_name=schema_name,
        table_names=table_names,
        column_names=column_names,
        include_table_names=include_table_names,
    )


def get_create_table_from_tables(
    *,
    table_name: str,
    tables: list,
    columns: list,
    join_clauses: list,
    table_aliases: list | None = None,
    column_aliases: dict | None = None,
    distinct: bool = False,
) -> str:
    """Generates a create table query from tables in the database

    :keyword table_name: the name of the resulting table
    :keyword tables: A list of length two for the source tables for the join
    :keyword columns: a list of columns to add to the table. Prepend with the table
        or the alias.
    :keyword join_clauses: a list of valid SQL join statements
    :keyword table_aliases: A list of length two of aliases for the source tables.
        ["a","b"] used if not supplied
    :keyword column_aliases: A dict of aliases to apply, like {'a_col': 'alias'}
    """
    if table_aliases is None:
        table_aliases = ["a", "b"]
    if column_aliases is None:
        column_aliases = {}
    for list_arg in [tables, table_aliases]:
        if len(list_arg) != 2:
            raise errors.CumulusLibraryError(
                f"Error generating table {table_name}: "
                f"Expected {','.join(list_arg)} to be of length two."
            )
    return get_base_template(
        "create_table_from_tables",
        table_name=table_name,
        join_tables=tables,
        join_table_aliases=table_aliases,
        join_columns=columns,
        column_aliases=column_aliases,
        join_clauses=join_clauses,
        distinct=distinct,
    )


def get_create_table_from_union(*, table_name: str, tables: list[str], columns: list[str]) -> str:
    """Generates a create union table query

    :keyword table_name: the name of the resulting table
    :keyword tables: A list of length two for the source tables for the join
    :keyword columns: a list of cols to add to the view. Prepend with the table
        or the alias."""
    return get_base_template(
        "create_table_from_union",
        table_name=table_name,
        tables=tables,
        columns=columns,
    )


def get_create_view_query(view_name: str, dataset: list[list[str]], view_cols: list[str]) -> str:
    """Generates a create view as query for inserting static data into athena

    :param view_name: The name of the athena table to create
    :param dataset: Array of data arrays to insert, i.e. [['1','3'],['2','4']]
    :param table_cols: Comma deleniated column names, i.e. ['first,second']
    """
    return get_base_template(
        "create_view_as",
        view_name=view_name,
        dataset=dataset,
        view_cols=view_cols,
    )


def get_create_view_from_tables(
    *,
    view_name: str,
    tables: list,
    columns: list,
    join_clauses: list,
    table_aliases: list | None = None,
    column_aliases: dict | None = None,
    distinct: bool = False,
) -> str:
    """Generates a create view query from tables in the database

    :keyword view_name: the name of the resulting view
    :keyword tables: A list of length two for the source tables for the join
    :keyword columns: a list of columns to add to the view. Prepend with the table
        or the alias.
    :keyword join_clauses: a list of valid SQL join statements
    :keyword table_aliases: A list of length two of aliases for the source tables.
        ["a","b"] used if not supplied
    :keyword column_aliases: A dict of aliases to apply, like {'a_col': 'alias'}
    """
    if table_aliases is None:
        table_aliases = ["a", "b"]
    if column_aliases is None:
        column_aliases = {}
    for list_arg in [tables, table_aliases]:
        if len(list_arg) != 2:
            raise errors.CumulusLibraryError(
                f"Error generating view {view_name}: "
                f"Expected {','.join(list_arg)} to be of length two."
            )
    return get_base_template(
        "create_view_from_tables",
        view_name=view_name,
        join_tables=tables,
        join_table_aliases=table_aliases,
        join_columns=columns,
        column_aliases=column_aliases,
        join_clauses=join_clauses,
        distinct=distinct,
    )


def get_ctas_from_parquet_query(
    schema_name: str,
    table_name: str,
    local_location: str,
    remote_location: str,
    table_cols: list[str],
    remote_table_cols_types: list[str],
) -> str:
    """Generates a create table as query using a directory of parquet files as a source

    This function will generate an appropriate query for the underlying DB.
    Not all params are used by each type of database.

    :param schema_name: (athena) The schema to create the table in
    :param table_name: (all) The name of the athena table to create
    :param local_location: (duckdb) A directory containing parquet files to group
        into a table
    :param remote_location: (athena) An S3 URL to a directory containing parquert
        fiels to group into a table
    :param table_cols: (all) names of fields in your parquet to use as SQL columns
    :param remote_table_cols_types: (athena) The types to assign to the columns
        created in athena. Note that these should not be SQL types, but instead
        should be parquet types.
    """
    return get_base_template(
        "ctas_from_parquet",
        schema_name=schema_name,
        table_name=table_name,
        local_location=local_location,
        remote_location=remote_location,
        table_cols=table_cols,
        remote_table_cols_types=remote_table_cols_types,
    )


def get_ctas_query(
    schema_name: str, table_name: str, dataset: list[list[str]], table_cols: list[str]
) -> str:
    """Generates a create table as query for inserting static data into athena

    Note that unlike other queries, the nature of the CTAS implementation in athena
    requires a schema name. This schema name should match the schema of your cursor,
    or the other queries in this template will not function correctly. All columns
    will be specified as varchar type.

    :param schema_name: The athena schema to create the table in
    :param table_name: The name of the athena table to create
    :param dataset: Array of data arrays to insert, i.e. [['1','3'],['2','4']]
    :param table_cols: Comma deleniated column names, i.e. ['first,second']
    """
    return get_base_template(
        "ctas",
        schema_name=schema_name,
        table_name=table_name,
        dataset=dataset,
        table_cols=table_cols,
    )


def get_ctas_query_from_df(schema_name: str, table_name: str, df: pandas.DataFrame) -> str:
    """Generates a create table as query from a dataframe

    :param schema_name: The athena schema to create the table in
    :param table_name: The name of the athena table to create
    :param df: A pandas dataframe
    """
    split_dict = df.to_dict(orient="split")
    return get_base_template(
        "ctas",
        schema_name=schema_name,
        table_name=table_name,
        dataset=split_dict["data"],
        table_cols=split_dict["columns"],
    )


def get_ctas_empty_query(
    schema_name: str,
    table_name: str,
    table_cols: list[str],
    table_cols_types: list[str] | None = None,
) -> str:
    """Generates a create table as query for initializing an empty table

    Note that unlike other queries, the nature of the CTAS implementation in athena
    requires a schema name. This schema name should match the schema of your cursor,
    or the other queries in this template will not function correctly. All columns
    will be specified as varchar type.

    :param schema_name: The athena schema to create the table in
    :param table_name: The name of the athena table to create
    :param table_cols: Comma deleniated column names, i.e. ['first,second']
    :param table_cols_types: Allows specifying a data type per column
      (default: all varchar)
    """
    if not table_cols_types:
        table_cols_types = ["varchar"] * len(table_cols)
    return get_base_template(
        "ctas_empty",
        schema_name=schema_name,
        table_name=table_name,
        table_cols=table_cols,
        table_cols_types=table_cols_types,
    )


def get_drop_view_table(name: str, view_or_table: str) -> str:
    """Generates a drop table if exists query"""
    if view_or_table in [e.value for e in TableView]:
        return get_base_template(
            "drop_view_table", view_or_table_name=name, view_or_table=view_or_table
        )


def get_extension_denormalize_query(config: sql_utils.ExtensionConfig) -> str:
    """extracts target extension from a table into a denormalized table

    This function is targeted at a complex extension element that is at the root
    of a FHIR resource - as an example, see the 5 codes at the root node of
    http://hl7.org/fhir/us/core/STU6/StructureDefinition-us-core-patient.html.
    The template will create a new table with the extension data, in arrays,
    mapped 1-1 to the table id. You can specify multiple systems
    in the ExtensionConfig passed to this function. For each patient, we'll
    take the data from the first extension coding system we find for each patient.

    :param config: An instance of ExtensionConfig.
    """
    return get_base_template(
        "extension_denormalize",
        source_table=config.source_table,
        source_id=config.source_id,
        target_table=config.target_table,
        target_col_prefix=config.target_col_prefix,
        fhir_extension=config.fhir_extension,
        ext_systems=config.ext_systems,
        is_array=config.is_array,
    )


def get_insert_into_query(
    schema: str,
    table_name: str,
    table_cols: list[str],
    dataset: list[list[str]],
    type_casts: dict | None = None,
) -> str:
    """Generates an insert query for adding data to an existing athena table

    :param schema_name: The athena query to create the table in
    :param table_name: The name of the athena table to create
    :param table_cols: Comma deleniated column names, i.e. ['first','second']
    :param dataset: Array of data arrays to insert, i.e. [['1','3'],['2','4']]
    """
    type_casts = type_casts or {}
    return get_base_template(
        "insert_into",
        schema_name=schema,
        table_name=table_name,
        table_cols=table_cols,
        dataset=dataset,
        type_casts=type_casts,
    )


def get_is_table_not_empty_query(
    source_table: str,
    field: str,
    unnests: list[dict] | None = None,
    conditions: list[str] | None = None,
):
    unnests = unnests or []
    conditions = conditions or []
    return get_base_template(
        "is_table_not_empty",
        source_table=source_table,
        field=field,
        unnests=unnests,
        conditions=conditions,
    )


def get_select_all_query(source_table: str):
    return get_base_template(
        "select_all",
        source_table=source_table,
    )


def get_show_tables(schema_name: str, prefix: str) -> str:
    """Generates a show tables query, filtered by prefix

    The intended use case for this function is to get a list of manifest study
    tables, so that they can be individually dropped during a clean call.

    :param schema_name: The athena schema to query
    :param table_name: The prefix to filter by. Jinja template auto adds '__'.
    """
    return get_base_template("show_tables", schema_name=schema_name, prefix=prefix)


def get_show_views(schema_name: str, prefix: str) -> str:
    """Generates a show vies query, filtered by prefix

    The intended use case for this function is to get a list of manifest study
    views, so that they can be individually dropped during a clean call.

    :param schema_name: The athena schema to query
    :param table_name: The prefix to filter by. Jinja template auto adds '__'.
    """
    return get_base_template("show_views", schema_name=schema_name, prefix=prefix)



"""Base Abstraction layer for interfacing with databases

By convention, to maintain this as a relatively light wrapper layer, if you have
to chose between a convenience function in a specific library (as an example, the
[pyathena to_sql function](https://github.com/laughingman7743/PyAthena/#to-sql))
or using raw sql directly in some form, you should do the latter. This not a law;
if there's a compelling reason to do so, just make sure you add an appropriate
wrapper method in one of DatabaseCursor or DatabaseBackend.
"""

import abc
import collections
import pathlib
from typing import Any, Protocol

import pandas


class DatabaseCursor(Protocol):
    """Protocol for a PEP-249 compatible cursor"""

    def execute(self, sql: str) -> None:
        pass  # pragma: no cover

    def fetchone(self) -> list | None:
        pass  # pragma: no cover

    def fetchmany(self, size: int | None) -> list[list] | None:
        pass  # pragma: no cover

    def fetchall(self) -> list[list] | None:
        pass  # pragma: no cover


class DatabaseParser(abc.ABC):
    """Parses information_schema results from a database"""

    @abc.abstractmethod
    def parse_found_schema(self, schema: dict[str, str]) -> dict:
        """Parses a database-provided schema string.

        :param schema: the results of a query from the get_column_datatype method
        of the template_sql.templates function. It looks like this (for athena at
        least, but the values are opaque strings and database-provider-specific):
            {
                'object_col': 'row(member_a varchar, member_b date)',
                'primitive_col': 'varchar',
            }

        :returns: a dictionary with an entry for every field present in the schema.
        For the above example, this should return:
            {
                'object_col': {
                    'member_a': {},
                    'member_b': {},
                },
                'primitive_col': {},
            }
        """

    def _recursively_validate(
        self, expected: dict[str, Any], schema: dict[str, Any]
    ) -> dict[str, Any]:
        schema = schema or {}
        output = {}
        for column, fields in expected.items():
            col_schema = schema.get(column.lower())

            # Is `fields` an falsy? (like: "recordedDate": None or [] or {})
            # This means we just want to check existance of `column`
            # otherwise this is a primitive col
            if not fields:
                output[column] = col_schema is not None

            # Is `fields` a list? (like: "subject": ["reference"])
            # This means we should check existance of all mentioned children.
            elif isinstance(fields, list):
                for field in fields:
                    subschema = self._recursively_validate({field: None}, col_schema)
                    output.setdefault(column, {}).update(subschema)

            # Is `fields` a dict?
            # Like: "component": {"valueQuantity": ["unit", "value"]}
            # This means we should descend one level
            elif isinstance(fields, dict):
                subschema = self._recursively_validate(fields, col_schema)
                output[column] = subschema

            else:
                raise ValueError("Bad expected schema provided")

        return output

    def validate_table_schema(self, expected: dict[str, list], schema: list[tuple]) -> dict:
        """Public interface for investigating if fields are in a table schema.

        expected is a dictionary of string column names to *something*:
        - falsy (like None or []): just check that the column exists
        - list of strings: check all the column children exist
        - dict of a new child 'expected' dictionary, with same above rules

        This method should lightly format results and pass them to
        parse_found_schema(), or a more bespoke table analysis function if needed.
        """
        parsed_schema = self.parse_found_schema(dict(schema))
        return self._recursively_validate(expected, parsed_schema)


class DatabaseBackend(abc.ABC):
    """A generic database backend, supporting basic cursor operations"""

    def __init__(self, schema_name: str):
        """Create connection to a database backend

        :param schema_name: the database name ('schema' is Athena-speak for a database)
        """
        self.schema_name = schema_name
        # db_type, while perhaps feeling redundant, is intended to be a value that is
        # passed to jinja templates for creating valid sql for a particular database's
        # technology
        self.db_type = None

    @abc.abstractmethod
    def cursor(self) -> DatabaseCursor:
        """Returns a connection to the backing database"""

    @abc.abstractmethod
    def pandas_cursor(self) -> DatabaseCursor:
        """Returns a connection to the backing database optimized for dataframes

        If your database does not provide an optimized cursor, this should function the
        same as a vanilla cursor.
        """

    @abc.abstractmethod
    def execute_as_pandas(
        self, sql: str, chunksize: int | None = None
    ) -> (pandas.DataFrame | collections.abc.Iterator[pandas.DataFrame], list[tuple]):
        """Returns a pandas.DataFrame version of the results from the provided SQL"""

    @abc.abstractmethod
    def parser(self) -> DatabaseParser:
        """Returns parser object for interrogating DB schemas"""

    def operational_errors(self) -> tuple[Exception]:
        """Returns a tuple of operational exception classes

        An operational error is something that went wrong while performing a database
        query. So something like "table doesn't exist" but not like a network or
        syntax error.

        This is designed to be used in an `except` clause.
        """
        return ()  # pragma: no cover

    def col_parquet_types_from_pandas(self, field_types: list) -> list:
        """Returns appropriate types for creating tables based from parquet.

        By default, returns an empty list (which assumes that the DB infers directly
        from parquet data types). Only override if your DB uses an explicit SerDe
        format, or otherwise needs a modified typing to inject directly into a query."""

        # The following example shows the types we're expecting to catch with this
        # approach and the rough type to cast them to.
        # TODO: consider handling complex types.
        # output = []
        # for field in field_types:
        #     match field:
        #         case numpy.dtypes.ObjectDType():
        #             output.append('string')
        #         case pandas.core.arrays.integer.Int64Dtype():
        #             output.append('int')
        #         case numpy.dtypes.Float64DType():
        #             output.append('float')
        #         case numpy.dtypes.BoolDType():
        #             output.append('bool')
        #         case numpy.dtypes.DateTime64DType():
        #             output.append('date')
        #         case _:
        #             raise errors.CumulusLibraryError(
        #                 f"Unsupported type {type(field)} found."
        #             )
        return []

    def col_pyarrow_types_from_sql(self, columns: list[tuple]) -> list:
        return columns  # pragma: no cover

    def upload_file(
        self,
        *,
        file: pathlib.Path,
        study: str,
        topic: str,
        remote_filename: str | None = None,
        force_upload=False,
    ) -> str | None:
        """Handler for remote database file upload.

        By default, this should return None. Only override this for databases that
        have an API for file upload (i.e. cloud databases)"""
        return None

    @abc.abstractmethod
    def create_schema(self, schema_name):
        """Creates a new schema object inside the catalog"""

    @abc.abstractmethod
    def close(self) -> None:
        """Clean up any resources necessary"""



"""A set of convenience functions for database logging"""

from cumulus_library import (
    __version__,
    base_utils,
    databases,
    enums,
    errors,
    study_manifest,
)
from cumulus_library.template_sql import base_templates, sql_utils


def log_transaction(
    *,
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    status: enums.LogStatuses | str | None = enums.LogStatuses.INFO,
    message: str | None = None,
):
    if isinstance(status, str):
        try:
            status = enums.LogStatuses(status)
        except ValueError as e:
            raise errors.CumulusLibraryError(
                f"Invalid event type {status} requested for transaction log.\n"
                f"Valid types: {','.join([x.value for x in enums.LogStatuses])}"
            ) from e
    _log_table(
        table=sql_utils.TransactionsTable(),
        config=config,
        manifest=manifest,
        dataset=[
            [
                manifest.get_study_prefix(),
                __version__,
                status.value,
                base_utils.get_utc_datetime(),
                message or None,
            ]
        ],
    )


def log_statistics(
    *,
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    table_type: str,
    table_name: str,
    view_name: str,
):
    _log_table(
        table=sql_utils.StatisticsTable(),
        config=config,
        manifest=manifest,
        dataset=[
            [
                manifest.get_study_prefix(),
                __version__,
                table_type,
                table_name,
                view_name,
                base_utils.get_utc_datetime(),
            ]
        ],
    )


def _log_table(
    *,
    table: sql_utils.BaseTable,
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    dataset: list[list],
):
    if manifest and manifest.get_dedicated_schema():
        db_schema = manifest.get_dedicated_schema()
        table_name = table.name
    else:
        db_schema = config.schema
        table_name = f"{manifest.get_study_prefix()}__{table.name}"
    query = base_templates.get_insert_into_query(
        schema=db_schema,
        table_name=table_name,
        table_cols=table.columns,
        dataset=dataset,
        type_casts=table.type_casts,
    )
    cursor = config.db.cursor()
    try:
        cursor.execute(query)
    except config.db.operational_errors() as e:
        # Migrating logging tables
        if "lib_transactions" in table_name:
            cols = cursor.execute(
                "SELECT column_name FROM information_schema.columns "  # noqa: S608
                f"WHERE table_name ='{table_name}' "
                f"AND table_schema ='{db_schema}'"
            ).fetchall()
            cols = [col[0] for col in cols]
            # Table schema pre-v3 library release
            if sorted(cols) == [
                "event_time",
                "library_version",
                "status",
                "study_name",
            ]:
                alter_query = ""
                if isinstance(config.db, databases.AthenaDatabaseBackend):
                    alter_query = (
                        f"ALTER TABLE {db_schema}.{table_name} " "ADD COLUMNS(message string)"
                    )
                elif isinstance(config.db, databases.DuckDatabaseBackend):
                    alter_query = (
                        f"ALTER TABLE {db_schema}.{table_name} " "ADD COLUMN message varchar"
                    )
                cursor.execute(alter_query)
                cursor.execute(query)
        else:
            raise e
